#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2008-2011 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import sys
import os
from os.path import abspath, dirname, isfile, isdir, basename, normpath
sys.path.insert(0, os.path.join(abspath(dirname(__file__)), 'lib'))
import shutil
import time
from skoollib import get_skool_class
from skoolhtml import get_skool_writer_class, join, FileInfo, SkoolWriter, SKOOLKIT_VERSION
from skoolparser import RefParser

SD_REF = 'sd.ref'
BTS_REF = 'bts.ref'
CHANGELOG_REF = 'changelog.ref'

def error(msg):
    sys.stderr.write('ERROR: %s\n' % msg)
    sys,exit(1)

def notify(notice):
    if verbose:
        print notice

def clock(operation, prefix, *args):
    if verbose:
        sys.stdout.write('%s ' % prefix)
        sys.stdout.flush()
        go = time.time()
    result = operation(*args)
    if verbose:
        sys.stdout.write('(%0.2fs)\n' % (time.time() - go))
    return result

def find(fname):
    if fname is None:
        return
    if isfile(fname):
        return fname
    fname = join('src', fname)
    if isfile(fname):
        return fname

def parse_args(args):
    verbose = True
    infiles = ((SD_REF, None), (BTS_REF, None))
    files = ''
    output_dir = None
    all_files = 'bCcdGghikLmopty'
    i = 0
    while i < len(args):
        arg = args[i]
        if arg == '-V':
            sys.stdout.write('SkoolKit %s\n' % SKOOLKIT_VERSION)
            sys.exit(0)
        if arg == '-q':
            verbose = False
        elif arg == '-b':
            infiles = ((BTS_REF, None),)
        elif arg == '-s':
            infiles = ((SD_REF, None),)
        elif arg == '-r':
            infiles = ((args[i + 1], None),)
            i += 1
        elif arg == '-f':
            infiles = ((None, args[i + 1]),)
            i += 1
        elif arg[:2] == '-w':
            files += arg[2:]
        else:
            output_dir = arg
        i += 1
    return verbose, infiles, files or all_files, output_dir

#
# Begin...
#
verbose, infiles, files, output_dir = parse_args(sys.argv[1:])
if output_dir is None:
    sys.stderr.write("Usage: %s [options] <output-dir>\n\n" % sys.argv[0])
    sys.stderr.write("Options:\n")
    sys.stderr.write("  -V       Show SkoolKit version number and exit\n")
    sys.stderr.write("  -q       Be quiet\n")
    sys.stderr.write("  -s       Write Skool Daze files only\n")
    sys.stderr.write("  -b       Write Back to Skool files only\n")
    sys.stderr.write("  -r FILE  Use this ref file\n")
    sys.stderr.write("  -f FILE  Use this skool file\n")
    sys.stderr.write("  -wX      Write only these files, where X is one or more of:\n")
    sys.stderr.write("             b = Bugs\n")
    sys.stderr.write("             C = Changelog\n")
    sys.stderr.write("             c = Character buffers\n")
    sys.stderr.write("             d = Disassembly files\n")
    sys.stderr.write("             G = Game status buffer\n")
    sys.stderr.write("             g = Graphics files\n")
    sys.stderr.write("             h = Home page\n")
    sys.stderr.write("             i = Disassembly index\n")
    sys.stderr.write("             k = Keypress table\n")
    sys.stderr.write("             L = Lessons and personal timetable index\n")
    sys.stderr.write("             m = Memory maps\n")
    sys.stderr.write("             o = Other code\n")
    sys.stderr.write("             p = Pokes\n")
    sys.stderr.write("             t = Trivia\n")
    sys.stderr.write("             y = Glossary\n")
    sys.exit(1)

pdir, bdir = os.path.split(output_dir)
if not bdir:
    pdir, bdir = os.path.split(pdir)
topdir = join(pdir, bdir)

# Create the top-level directory if necessary
if not isdir(topdir):
    notify('Creating directory %s' % topdir)
    os.makedirs(topdir)

if not files.replace('C', '').replace('h', ''):
    # Don't parse a skool file if just writing the changelog or the home page
    modes = ''

skools = infiles and infiles[0][0] in (SD_REF, BTS_REF)

for reffile, skoolfile in infiles:
    ref_parser = RefParser()
    reffile_f = find(reffile)
    config = {}

    if reffile:
        if reffile_f is None:
            error('%s: ref file not found' % reffile)
        ref_parser.parse(reffile_f)
        config = ref_parser.build_dictionary('Config')
        prefix = basename(reffile)
        if prefix[-4:] == '.ref':
            prefix = prefix[:-4]
        skoolfile = config.get('SkoolFile', '%s.skool' % prefix)

    skoolfile_f = find(skoolfile)
    if skoolfile_f is None:
        error('%s: skool file not found' % skoolfile)

    prefix = basename(skoolfile_f)
    if prefix[-6:] == '.skool':
        prefix = prefix[:-6]

    if reffile is None:
        reffile_f = find('%s.ref' % prefix)
        if reffile_f:
            ref_parser.parse(reffile_f)
            config = ref_parser.build_dictionary('Config')

    if skoolfile:
        if reffile_f:
            notify('Using ref file: %s' % reffile_f)
        else:
            notify('Found no ref file for %s' % skoolfile_f)
    elif reffile:
        notify('Using skool file: %s' % skoolfile_f)

    skool_class = get_skool_class(config.get('SkoolClass'))
    skool_writer_class = get_skool_writer_class(config.get('SkoolWriterClass'))
    game_dir = config.get('GameDir', prefix)

    # Parse the skool file and initialise the writer
    skool = clock(skool_class, 'Parsing %s' % skoolfile_f, skoolfile_f, ref_parser)
    file_info = FileInfo(topdir, game_dir)
    skool_writer = skool_writer_class(skool, file_info)
    paths = skool_writer.paths

    # Create the disassembly subdirectory if necessary
    odir = file_info.odir
    if not isdir(odir):
        notify('Creating directory %s' % odir)
        os.makedirs(odir)

    # Copy style.css and skool.js if necessary
    for fname, dest in ((paths.get('StyleSheet'), paths.get('StyleSheetPath', '')), (paths.get('JavaScript'), paths.get('JavaScriptPath', ''))):
        if fname:
            f = find(fname)
            if f:
                html_f = normpath(join(topdir, game_dir, dest, fname))
                if not isfile(html_f) or os.stat(f).st_mtime > os.stat(html_f).st_mtime:
                    notify('Copying %s to %s' % (f, html_f))
                    shutil.copy2(f, html_f)
            else:
                error('<%s>: file not found' % fname)

    # Write logo image file if necessary
    if not skool_writer.file_exists(paths['Logo']) and skool_writer.write_logo_image(odir):
        notify('  Wrote %s' % join(game_dir, paths['Logo']))

    # Write disassembly files
    if 'd' in files:
        clock(skool_writer.write_asm_entries, '  Writing disassembly files in %s' % join(game_dir, skool.code_path))

    # Write the memory map files
    if 'm' in files:
        clock(skool_writer.write_memory_map, '  Writing %s' % join(game_dir, paths['MemoryMap']))
        clock(skool_writer.write_routines_map, '  Writing %s' % join(game_dir, paths['RoutinesMap']))
        clock(skool_writer.write_data_map, '  Writing %s' % join(game_dir, paths['DataMap']))
        clock(skool_writer.write_messages_map, '  Writing %s' % join(game_dir, paths['MessagesMap']))

    # Write graphics files
    if 'g' in files:
        if skool_writer.has_feature('PlayArea'):
            clock(skool_writer.write_playarea, '  Writing %s' % join(game_dir, paths['PlayArea']))
        if skool_writer.has_feature('PlayAreaTiles'):
            clock(skool_writer.write_playarea_tiles, '  Writing %s' % join(game_dir, paths['PlayAreaTiles']))
        if skool_writer.has_feature('InitialAnimatoryStates'):
            clock(skool_writer.write_asstart, '  Writing %s' % join(game_dir, paths['InitialAnimatoryStates']))
        if skool_writer.has_feature('AnimatoryStates'):
            clock(skool_writer.write_animatory_states, '  Writing %s' % join(game_dir, paths['AnimatoryStates']))
        if skool_writer.has_feature('AnimatoryStateTiles'):
            clock(skool_writer.write_animatory_state_tiles, '  Writing %s' % join(game_dir, paths['AnimatoryStateTiles']))
        if skool_writer.has_graphics():
            clock(skool_writer.write_graphics, '  Writing %s' % join(game_dir, paths['Graphics']))
        if skool_writer.has_graphic_glitches():
            clock(skool_writer.write_graphic_glitches, '  Writing %s' % join(game_dir, paths['GraphicGlitches']))

    # Write lesson files and personal timetables index
    if 'L' in files and skool_writer.has_feature('Lessons'):
        clock(skool_writer.write_lessons, '  Writing lesson files in %s' % join(game_dir, paths['LessonsPath']))
        clock(skool_writer.write_lesson_index, '  Writing %s' % join(game_dir, paths['LessonIndex']))
        clock(skool_writer.write_personal_timetable_index, '  Writing %s' % join(game_dir, paths['PersonalTimetables']))

    # Write other files
    if 'G' in files and skool_writer.has_gbuffer():
        clock(skool_writer.write_gbuffer, '  Writing %s' % join(game_dir, paths['GameStatusBuffer']))
    if 'b' in files and skool_writer.has_bugs():
        clock(skool_writer.write_bugs, '  Writing %s' % join(game_dir, paths['Bugs']))
    if 't' in files and skool_writer.has_facts():
        clock(skool_writer.write_facts, '  Writing %s' % join(game_dir, paths['Facts']))
    if 'y' in files and skool_writer.has_glossary():
        clock(skool_writer.write_glossary, '  Writing %s' % join(game_dir, paths['Glossary']))
    if 'p' in files and skool_writer.has_pokes():
        clock(skool_writer.write_pokes, '  Writing %s' % join(game_dir, paths['Pokes']))
    if 'c' in files and skool_writer.has_feature('CharacterBuffers'):
        clock(skool_writer.write_character_buffers, '  Writing %s' % join(game_dir, paths['CharacterBuffers']))
    if 'k' in files and skool_writer.has_feature('KeypressTable'):
        clock(skool_writer.write_keytable, '  Writing %s' % join(game_dir, paths['Keys']))
    if 'o' in files:
        for code_id, code in skool.other_code:
            skoolfile = find(code['Source'])
            skool2 = clock(skool_class, '  Parsing %s' % skoolfile, skoolfile, ref_parser, code_id)
            skool2_writer = skool_writer_class(skool2, file_info)
            path = code['Path']
            index = code['Index']
            clock(skool2_writer.write_map, '    Writing %s' % join(game_dir, index), index, code['Title'], '', path)
            clock(skool2_writer.write_entries, '    Writing disassembly files in %s' % join(game_dir, path), path, index, code['Header'])

    # Write index.html
    if 'i' in files:
        clock(skool_writer.write_index, '  Writing %s' % join(game_dir, paths['GameIndex']))

if skools:
    # Copy style.css if necessary
    css = 'style.css'
    f = find(css)
    if f:
        html_f = join(topdir, css)
        if not isfile(html_f) or os.stat(f).st_mtime > os.stat(html_f).st_mtime:
            notify('Copying %s to %s' % (f, html_f))
            shutil.copy2(f, topdir)
    changelog = 'changelog.html'
    if 'h' in files:
        sd_ref_parser = RefParser()
        sd_ref_parser.parse(find(SD_REF))
        sd_config = sd_ref_parser.build_dictionary('Config')
        sd_dir = sd_config['GameDir']
        sd_paths = sd_ref_parser.build_dictionary('Paths')
        sd_index = sd_paths.get('GameIndex', 'index.html')
        bts_ref_parser = RefParser()
        bts_ref_parser.parse(find(BTS_REF))
        bts_config = bts_ref_parser.build_dictionary('Config')
        bts_dir = bts_config['GameDir']
        bts_paths = bts_ref_parser.build_dictionary('Paths')
        bts_index = bts_paths.get('GameIndex', 'index.html')
        home = join(topdir, 'index.html')
        games = (('Skool Daze', sd_dir, sd_index), ('Back to Skool', bts_dir, bts_index))
        clock(SkoolWriter.write_home_page, 'Writing %s' % home, home, changelog, css, games)
    if 'C' in files:
        clock(SkoolWriter.write_changelog, 'Writing %s' % join(topdir, changelog), topdir, changelog, find(CHANGELOG_REF), css)
